/*======================================================================

  Device driver for the PCMCIA control functionality of EP93xx
  microprocessors.  Derived from pxa2xx_base.c

    The contents of this file may be used under the
    terms of the GNU Public License version 2 (the "GPL")

    (c) Ian Molton (spyro@f2s.com) 2003
    (c) Stefan Eletzhofer (stefan.eletzhofer@inquant.de) 2003,4
    (c) Robert Whaley (rwhaley@applieddata.net) 2005

    derived from sa11xx_base.c

     Portions created by John G. Dorsey are
     Copyright (C) 1999 John G. Dorsey.

  ======================================================================*/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/config.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/spinlock.h>

#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>

#include <pcmcia/cs_types.h>
#include <pcmcia/ss.h>
#include <pcmcia/bulkmem.h>
#include <pcmcia/cistpl.h>

#include "cs_internal.h"
#include "soc_common.h"

#define MIN(a,b)	((a)>(b) ? (b) : (a))

#define MCXX_SETUP_MAX     0xff
#define MCXX_ASST_MAX      0xff
#define MCXX_HOLD_MAX      0x0f

#define MCXX_SETUP_SHIFT   0
#define MCXX_ASST_SHIFT    16
#define MCXX_HOLD_SHIFT    8



static inline u_int ep93xx_mcxx_hold(u_int pcmcia_cycle_ns,
				     u_int hclk_10khz)
{
	u_int code = pcmcia_cycle_ns * hclk_10khz;
	return (code / 100000) + ((code % 100000) ? 1 : 0) - 1;
}

static inline u_int ep93xx_mcxx_asst(u_int pcmcia_cycle_ns,
				     u_int hclk_10khz)
{
	u_int code = pcmcia_cycle_ns * hclk_10khz;
	return (code / 100000) + ((code % 100000) ? 1 : 0) - 1;
}

static inline u_int ep93xx_mcxx_setup(u_int pcmcia_cycle_ns,
				      u_int hclk_10khz)
{
	u_int code = pcmcia_cycle_ns * hclk_10khz;
	return (code / 100000) + ((code % 100000) ? 1 : 0) - 1;
}

static int ep93xx_pcmcia_set_mcmem( int sock, int speed, int clock )
{
        writel(PCCONFIG_MW_16BIT
	     | (MIN(ep93xx_mcxx_setup(speed, clock), MCXX_SETUP_MAX) << MCXX_SETUP_SHIFT)
	     | (MIN(ep93xx_mcxx_asst(speed, clock), MCXX_ASST_MAX) << MCXX_ASST_SHIFT)
	     | (MIN(ep93xx_mcxx_hold(speed, clock), MCXX_HOLD_MAX) << MCXX_HOLD_SHIFT),
	     SMC_PCCommon);

	return 0;
}

static int ep93xx_pcmcia_set_mcio( int sock, int speed, int clock )
{
        writel(PCCONFIG_MW_16BIT
	     | (MIN(ep93xx_mcxx_setup(speed, clock), MCXX_SETUP_MAX) << MCXX_SETUP_SHIFT)
	     | (MIN(ep93xx_mcxx_asst(speed, clock), MCXX_ASST_MAX) << MCXX_ASST_SHIFT)
	     | (MIN(ep93xx_mcxx_hold(speed, clock), MCXX_HOLD_MAX) << MCXX_HOLD_SHIFT),
	     SMC_PCIO);

	return 0;
}

static int ep93xx_pcmcia_set_mcatt( int sock, int speed, int clock )
{
        writel(PCCONFIG_MW_16BIT
	     | (MIN(ep93xx_mcxx_setup(speed, clock), MCXX_SETUP_MAX) << MCXX_SETUP_SHIFT)
	     | (MIN(ep93xx_mcxx_asst(speed, clock), MCXX_ASST_MAX) << MCXX_ASST_SHIFT)
	     | (MIN(ep93xx_mcxx_hold(speed, clock), MCXX_HOLD_MAX) << MCXX_HOLD_SHIFT),
	     SMC_PCAttribute);

	return 0;
}

static int ep93xx_pcmcia_set_mcxx(struct soc_pcmcia_socket *skt, unsigned int clk)
{
	struct soc_pcmcia_timing timing;
	int sock = skt->nr;

	soc_common_pcmcia_get_timing(skt, &timing);

	ep93xx_pcmcia_set_mcmem(sock, timing.mem, clk);
	ep93xx_pcmcia_set_mcatt(sock, timing.attr, clk);
	ep93xx_pcmcia_set_mcio(sock, timing.io, clk);

	return 0;
}

static int ep93xx_pcmcia_set_timing(struct soc_pcmcia_socket *skt)
{
	unsigned int hclk = 10000; /* for now just hard code hclk to 100Mhz (in 10khz units)*/
	return ep93xx_pcmcia_set_mcxx(skt, hclk);
}

int ep93xx_drv_pcmcia_probe(struct device *dev)
{
	int ret;
	struct pcmcia_low_level *ops;
	int first, nr;

	if (!dev || !dev->platform_data)
		return -ENODEV;

	ops = (struct pcmcia_low_level *)dev->platform_data;
	first = ops->first;
	nr = ops->nr;

	/* Provide our EP93xx specific timing routines. */
	ops->set_timing  = ep93xx_pcmcia_set_timing;

	ret = soc_common_drv_pcmcia_probe(dev, ops, first, nr);

	return ret;
}
EXPORT_SYMBOL(ep93xx_drv_pcmcia_probe);

static int ep93xx_drv_pcmcia_suspend(struct device *dev, u32 state, u32 level)
{
	int ret = 0;
	if (level == SUSPEND_SAVE_STATE)
		ret = pcmcia_socket_dev_suspend(dev, state);
	return ret;
}

static int ep93xx_drv_pcmcia_resume(struct device *dev, u32 level)
{
	int ret = 0;
	if (level == RESUME_RESTORE_STATE)
	{
		ret = pcmcia_socket_dev_resume(dev);
	}
	return ret;
}

static struct device_driver ep93xx_pcmcia_driver = {
	.probe		= ep93xx_drv_pcmcia_probe,
	.remove		= soc_common_drv_pcmcia_remove,
	.suspend 	= ep93xx_drv_pcmcia_suspend,
	.resume 	= ep93xx_drv_pcmcia_resume,
	.name		= "ep93xx-pcmcia",
	.bus		= &platform_bus_type,
};

static int __init ep93xx_pcmcia_init(void)
{
	return driver_register(&ep93xx_pcmcia_driver);
}

static void __exit ep93xx_pcmcia_exit(void)
{
	driver_unregister(&ep93xx_pcmcia_driver);
}

module_init(ep93xx_pcmcia_init);
module_exit(ep93xx_pcmcia_exit);

MODULE_AUTHOR("Robert Whaley (whaley@applieddata.net)");
MODULE_DESCRIPTION("Linux PCMCIA Card Services: EP93xx core socket driver");
MODULE_LICENSE("GPL");
